import os
import logging
from enum import Enum

from eth_utils import decode_hex, is_same_address
from typing import Callable

from .context import Context
from ..utils import (
    get_event_blocking,
    signed_contract_transaction,
    sign_balance_proof,
    verify_closing_sig,
    keccak256
)

logger = logging.getLogger('atn.microraiden' + os.path.splitext(os.path.basename(__file__))[0])

class Channel:
    class State(Enum):
        open = 1
        settling = 2
        closed = 3

    def __init__(
            self,
            core: Context,
            sender: str,
            receiver: str,
            block: int,
            deposit: int = 0,
            balance: int = 0,
            state: State = State.open,
            on_settle: Callable[['Channel'], None] = lambda channel: None
    ):
        self._balance = 0
        self._balance_sig = None

        self.core = core
        self.sender = sender
        self.receiver = receiver
        self.deposit = deposit
        self.block = block
        self.update_balance(balance)
        self.state = state
        self.on_settle = on_settle

        assert self.block is not None
        assert self._balance_sig

    @property
    def balance(self):
        return self._balance

    @property
    def key(self) -> bytes:
        return keccak256(self.sender, self.receiver, self.block)

    def update_balance(self, value):
        self._balance = value
        self._balance_sig = self.sign()

    @property
    def balance_sig(self):
        return self._balance_sig

    def sign(self):
        sig =  sign_balance_proof(
            self.core.private_key,
            self.receiver,
            self.block,
            self.balance,
            self.core.channel_manager.address
        )
        return sig

    def topup(self, deposit):
        """
        Attempts to increase the deposit in an existing channel. Block until confirmation.
        """
        if self.state != Channel.State.open:
            logger.error('Channel must be open to be topped up.')
            return

        _balance = self.core.web3.eth.getBalance(self.core.address)
        if _balance < deposit:
            logger.error(
                'Insufficient tokens available for the specified topup ({}/{})'
                .format(_balance, deposit)
            )
            return None

        logger.info('Topping up channel to {} created at block #{} by {} ATN.'.format(
            self.receiver, self.block, deposit
        ))
        current_block = self.core.web3.eth.blockNumber

        data = (decode_hex(self.sender) +
                decode_hex(self.receiver) +
                self.block.to_bytes(4, byteorder='big'))
        tx = signed_contract_transaction(
            self.core.account,
            self.core.channel_manager,
            'topUp(address,uint32)',
            [
                self.receiver,
                self.block
            ],
            deposit
        )
        self.core.web3.eth.sendRawTransaction(tx.rawTransaction)

        logger.debug('Waiting for topup confirmation event...')
        event = get_event_blocking(
            self.core.channel_manager,
            'ChannelToppedUp',
            from_block=current_block + 1,
            argument_filters={
                '_sender_address': self.sender,
                '_receiver_address': self.receiver,
                '_open_block_number': self.block
            }
        )

        if event:
            logger.debug('Successfully topped up channel in block {}.'.format(event['blockNumber']))
            self.deposit += deposit
            return event
        else:
            logger.error('No event received.')
            return None

    def close(self, balance=None):
        """
        Attempts to request close on a channel. An explicit balance can be given to override the
        locally stored balance signature. Blocks until a confirmation event is received or timeout.
        """
        if self.state != Channel.State.open:
            logger.error('Channel must be open to request a close.')
            return
        logger.info('Requesting close of channel to {} created at block #{}.'.format(
            self.receiver, self.block
        ))
        current_block = self.core.web3.eth.blockNumber

        if balance is not None:
            self.update_balance(balance)

        tx = signed_contract_transaction(
            self.core.account,
            self.core.channel_manager,
            'uncooperativeClose(address,uint32,uint256)',
            [
                self.receiver,
                self.block,
                self.balance
            ]
        )
        self.core.web3.eth.sendRawTransaction(tx.rawTransaction)

        logger.debug('Waiting for close confirmation event...')
        event = get_event_blocking(
            self.core.channel_manager,
            'ChannelCloseRequested',
            from_block=current_block + 1,
            argument_filters={
                '_sender_address': self.sender,
                '_receiver_address': self.receiver,
                '_open_block_number': self.block
            }
        )

        if event:
            logger.debug('Successfully sent channel close request in block {}.'.format(
                event['blockNumber']
            ))
            self.state = Channel.State.settling
            return event
        else:
            logger.error('No event received.')
            return None

    def close_cooperatively(self, closing_sig: bytes, contract_receiver_owner: str = None):
        """
        Attempts to close the channel immediately by providing a hash of the channel's balance
        proof signed by the receiver. This signature must correspond to the balance proof stored in
        the passed channel state.
        """
        if self.state == Channel.State.closed:
            logger.error('Channel must not be closed already to be closed cooperatively.')
            return None
        logger.info('Attempting to cooperatively close channel to {} created at block #{}.'.format(
            self.receiver, self.block
        ))
        current_block = self.core.web3.eth.blockNumber
        receiver_recovered = verify_closing_sig(
            self.sender,
            self.block,
            self.balance,
            closing_sig,
            self.core.channel_manager.address
        )
        bytecode = self.core.web3.eth.getCode(self.receiver)
        if bytecode:
            # The channel's receiver is a contract, the close_signature should be signed by it's owner
            assert(contract_receiver_owner is not None)
            if contract_receiver_owner is not None and not is_same_address(receiver_recovered, contract_receiver_owner):
                logger.error('Invalid closing signature')
                return None
        else:
            if not is_same_address(receiver_recovered, self.receiver):
                logger.error('Invalid closing signature.')
                return None

        tx = signed_contract_transaction(
            self.core.account,
            self.core.channel_manager,
            'cooperativeClose(address,uint32,uint256,bytes,bytes)',
            [
                self.receiver,
                self.block,
                self.balance,
                self.balance_sig,
                closing_sig
            ]
        )
        self.core.web3.eth.sendRawTransaction(tx.rawTransaction)

        logger.debug('Waiting for settle confirmation event...')
        event = get_event_blocking(
            self.core.channel_manager,
            'ChannelSettled',
            from_block=current_block + 1,
            argument_filters={
                '_sender_address': self.sender,
                '_receiver_address': self.receiver,
                '_open_block_number': self.block
            }
        )

        if event:
            logger.debug('Successfully closed channel in block {}.'.format(event['blockNumber']))
            self.state = Channel.State.closed
            return event
        else:
            logger.error('No event received.')
            return None

    def settle(self):
        """
        Attempts to settle a channel that has passed its settlement period. If a channel cannot be
        settled yet, the call is ignored with a warning. Blocks until a confirmation event is
        received or timeout.
        """
        if self.state != Channel.State.settling:
            logger.error('Channel must be in the settlement period to settle.')
            return None
        logger.info('Attempting to settle channel to {} created at block #{}.'.format(
            self.receiver, self.block
        ))

        _, _, settle_block, _, _ = self.core.channel_manager.call().getChannelInfo(
            self.sender, self.receiver, self.block
        )

        current_block = self.core.web3.eth.blockNumber
        wait_remaining = settle_block - current_block
        if wait_remaining > 0:
            logger.warning('{} more blocks until this channel can be settled. Aborting.'.format(
                wait_remaining
            ))
            return None

        tx = signed_contract_transaction(
            self.core.account,
            self.core.channel_manager,
            'settle(address,uint32)',
            [
                self.receiver,
                self.block
            ]
        )
        self.core.web3.eth.sendRawTransaction(tx.rawTransaction)

        logger.debug('Waiting for settle confirmation event...')
        event = get_event_blocking(
            self.core.channel_manager,
            'ChannelSettled',
            from_block=current_block + 1,
            argument_filters={
                '_sender_address': self.sender,
                '_receiver_address': self.receiver,
                '_open_block_number': self.block
            }
        )

        if event:
            logger.debug('Successfully settled channel in block {}.'.format(event['blockNumber']))
            self.state = Channel.State.closed
            self.on_settle(self)
            return event
        else:
            logger.error('No event received.')
            return None

    def create_transfer(self, value):
        """
        Updates the given channel's balance and balance signature with the new value. The signature
        is returned and stored in the channel state.
        """
        assert value >= 0
        if self.remain_balance() < value:
            logger.error(
                'Insufficient funds on channel. Needed: {}. Available: {}/{}.'
                .format(value, self.deposit - self.balance, self.deposit)
            )
            return None

        logger.debug('Signing new transfer of value {} on channel to {} created at block #{}.'.format(
            value, self.receiver, self.block
        ))

        if self.state == Channel.State.closed:
            logger.error('Channel must be open to create a transfer.')
            return None

        self.update_balance(self.balance + value)

        return self.balance_sig

    def is_valid(self) -> bool:
        return self.sign() == self.balance_sig and self.balance <= self.deposit

    def remain_balance(self) -> int:
        return self.deposit - self.balance

    def is_suitable(self, value: int):
        return self.remain_balance() >= value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyatn_client import Atn

# API info
DBOTADDRESS = '0xfd4F504F373f0af5Ff36D9fbe1050E6300699230' # address of the DBot you want to test, use 'AI poetry' as example
URI = '/reg'        # uri of the DBot's API endpoint which you want to call
METHOD = 'POST'     # method of the DBot's API endpoint which you want to call
requests_kwargs = {
    "data": {
        "theme": "中秋月更圆"
    }
}

# Init Atn
atn = Atn(
    pk_file='./dbot-examples/accounts/signer02/keystore/UTC--2018-06-19T08-46-10.312652809Z--0cc1f6e1a55b163301434a47b1cb68cbd7e27cad',
    pw_file='./dbot-examples/accounts/signer02/password/passwd'
)

# Call a DBot API 12 times
for i in range(12):
    response = atn.call_dbot_api(dbot_address=DBOTADDRESS,
                                uri=URI,
                                method=METHOD,
                                **requests_kwargs)
    print('Call {}:\n{}'.format(i + 1, response.text))

# close the channel only when you do not need it any more,
# the remain balance in the channel will be returned to your account
atn.close_channel(DBOTADDRESS)



class MicroRaidenException(Exception):
    """Base exception for uRaiden"""
    pass


class InvalidBalanceAmount(MicroRaidenException):
    """Raised if the payment contains lesser balance than the previous one."""
    pass


class InvalidBalanceProof(MicroRaidenException):
    """Balance proof data do not make sense."""
    pass


class NoOpenChannel(MicroRaidenException):
    """Attempt to use nonexisting channel."""
    pass


class InsufficientConfirmations(MicroRaidenException):
    """uRaiden channel doesn't have enough confirmations."""
    pass


class NoBalanceProofReceived(MicroRaidenException):
    """Attempt to close channel with no registered payments."""
    pass


class InvalidContractVersion(MicroRaidenException):
    """Library is not compatible with the deployed contract version"""
    pass


class StateFileException(MicroRaidenException):
    """Base exception class for state file (database) operations"""
    pass


class ReceiverAddrMismatchPrivateKey(MicroRaidenException):
    """Receiver address mismatch with private key."""
    pass

class StateContractAddrMismatch(StateFileException):
    """Stored state contract address doesn't match."""
    pass


class StateReceiverAddrMismatch(StateFileException):
    """Stored state receiver address doesn't match."""
    pass


class StateFileLocked(StateFileException):
    """Another process is already using the database"""
    pass


class InsecureStateFile(StateFileException):
    """Permissions of the state file do not match (0600 is expected)."""
    pass


class NetworkIdMismatch(StateFileException):
    """RPC endpoint and database have different network id."""
    pass



from web3.utils.filters import construct_event_filter_params
import functools
from web3.utils.events import get_event_data


class LogFilter:
    def __init__(
        self,
        web3,
        abi,
        address,
        event_name,
        from_block=0,
        to_block='latest',
        filters=None,
        callback=None
    ):
        self.web3 = web3
        self.event_name = event_name

        # Callback for every registered log
        self.callback = callback

        filter_kwargs = {
            'fromBlock': from_block,
            'toBlock': to_block,
            'address': address
        }

        event_abi = [i for i in abi if i['type'] == 'event' and i['name'] == event_name]
        if len(event_abi) == 0:
            return None

        self.event_abi = event_abi[0]
        assert self.event_abi

        filters = filters if filters else {}

        data_filter_set, filter_params = construct_event_filter_params(
            self.event_abi,
            argument_filters=filters,
            **filter_kwargs
        )
        log_data_extract_fn = functools.partial(get_event_data, event_abi)

        self.filter = web3.eth.filter(filter_params)
        self.filter.set_data_filters(data_filter_set)
        self.filter.log_entry_formatter = log_data_extract_fn
        self.filter.filter_params = filter_params

    def init(self, post_callback=None):
        for log in self.get_logs():
            log['event'] = self.event_name
            self.callback(log)
        if post_callback:
            post_callback()

    def get_logs(self):
        logs = self.web3.eth.getFilterLogs(self.filter.filter_id)
        formatted_logs = []
        for log in [dict(log) for log in logs]:
            formatted_logs.append(self.set_log_data(log))
        return formatted_logs

    def set_log_data(self, log):
        log['args'] = get_event_data(self.event_abi, log)['args']
        log['event'] = self.event_name
        return log

    def uninstall(self):
        assert self.web3 is not None
        assert self.filter is not None
        self.web3.eth.uninstallFilter(self.filter.filter_id)
        self.filter = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

def remove_slash_prefix(uri):
    if uri.startswith('/'):
        return uri[1:]
    else:
        return uri

def tobytes32(s):
    length = len(s.encode('utf-8'))
    assert(length <= 256)
    return  s.encode('utf-8') + b'\0' * (32 - length)



"""
This file contains configuration constants you probably don't need to change
"""
import json
import os

def read_version(path: str):
    return open(path, 'r').read().strip()

# api path prefix
API_PATH = "/api/1"
"""str: api path prefix"""

#  MICRORAIDEN_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
"""str: absolute path to module directory. Used to find path to the webUI sources"""
#  HTML_DIR = os.path.join(MICRORAIDEN_DIR, 'microraiden', 'webui')
"""str: webUI sources directory"""
#  JSLIB_DIR = os.path.join(HTML_DIR, 'js')
"""str: javascript directory"""
#  JSPREFIX_URL = '/js'
"""str: url prefix for jslib dir"""

WEB3_PROVIDER_DEFAULT = "http://0.0.0.0:8545"
"""str: ethereum node RPC interface URL"""

CHANNEL_MANAGER_NAME = 'TransferChannels'
"""str: name of the channel manager contract"""

CONTRACTS_JSON = 'contracts.json'
"""str: compiled contracts path"""

with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), CONTRACTS_JSON)) as fh:
    CONTRACT_METADATA = json.load(fh)

PROXY_BALANCE_LIMIT = 10**8
"""int: proxy will stop serving requests if receiver balance is below PROXY_BALANCE_LIMIT"""
SLEEP_RELOAD = 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import logging
import logging.config
import time
from typing import Callable, Tuple, Union
import json
from munch import Munch
import requests
from requests import Response

from web3 import Web3, HTTPProvider
from web3.middleware import geth_poa_middleware
from eth_utils import is_same_address, decode_hex, encode_hex

from .microraiden.header import HTTPHeaders
from .microraiden.client import Client, Channel
from .microraiden.utils import verify_balance_proof

from .utils import remove_slash_prefix, tobytes32
from .log import AtnLogger

logger = logging.getLogger('atn')

class AtnException(Exception):
    """Base exception for ATN Client"""
    pass

def _make_dbot_contract(web3, dbot_address):
    contracts_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'contracts.json')
    if os.path.exists(contracts_file):
        with open(contracts_file) as fh:
            contracts = json.load(fh)
    else:
        contracts = json.loads(pkg_resources.resource_string('pyatn_client', 'contracts.json'))
    dbot_address = Web3.toChecksumAddress(dbot_address)
    dbotContract = web3.eth.contract(address=dbot_address,
                                     abi=contracts['Dbot']['abi'],
                                     bytecode=contracts['Dbot']['bytecode'])
    return dbotContract

class Atn():
    """ATN Client Class

    Usage::

      >>> from pyatn_client import Atn
      >>> atn = Atn(pk_file='<path to keystore file>',
      >>>           pw_file='<path to password file>'
      >>> )
      >>> resp = atn.call_dbot_api(dbot_address='0xfd4F504F373f0af5Ff36D9fbe1050E6300699230',
      >>>                          uri='/reg',
      >>>                          method='POST',
      >>>                          data={'theme': '中秋月更圆'})
      <Response [200]>
    """
    def __init__(
        self,
        pk_file: str,
        pw_file: str,
        http_provider: str = 'https://rpc-test.atnio.net',
        deposit_strategy: Callable[[int], int] = lambda value: 10 * value
    ) -> None:
        """Init Atn Class

        """
        logging.config.dictConfig(AtnLogger('atn').config())

        w3 = Web3(HTTPProvider(http_provider))
        w3.middleware_stack.inject(geth_poa_middleware, layer=0)

        self.deposit_strategy = deposit_strategy
        self.channel_client = Client(
            private_key=pk_file,
            key_password_path=pw_file,
            web3=w3
        )

    def set_deposit_strategy(self, deposit_strategy: Callable[[int], int]) -> None:
        """Change deposit strategy.

        Channel will be auto created if no channel or be topuped if
        insufficient balance in channel when `call_dbot_api` be called.
        The deposit value is determined by `deposit_strategy`.

        :param deposit_strategy: callable function to determine the deposit
            value when create and topup channel.
            If it's `None`, disable auto create or topup channel
        """
        self.deposit_strategy = deposit_strategy

    def get_dbot_name(self, dbot_address: str) -> str:
        """Get the name of DBot according the address of DBot contract
        on ATN blockchain.

        :param dbot_address: address of the DBot contract
        :return: name of the DBot
        :rtype: str
        """
        dbot_address = Web3.toChecksumAddress(dbot_address)
        w3 = self.channel_client.context.web3
        Dbot = _make_dbot_contract(w3, dbot_address)
        name = Dbot.functions.name().call()
        return name.decode('utf-8').rstrip('\0')

    def get_dbot_domain(self, dbot_address: str) -> str:
        """Get the domain of DBot according the address of DBot contract
        on ATN blockchain.

        The DBot server should be accessed on the domain.
        The domain may contain `http://` or `https://` protocol prefix.

        :param dbot_address: address of the DBot contract
        :return: domain of the DBot
        :rtype: str
        """
        dbot_address = Web3.toChecksumAddress(dbot_address)
        w3 = self.channel_client.context.web3
        Dbot = _make_dbot_contract(w3, dbot_address)
        domain = Dbot.functions.domain().call()
        return domain.decode('utf-8').rstrip('\0')

    def get_dbot_owner(self, dbot_address: str) -> str:
        """Get the owner account of DBot contract on ATN blockchain.

        Close signature should be signed by the owner of DBot
        when DBot user want to cooperative close the channel with DBot.

        :param dbot_address: address of the DBot contract
        :return: account address of owner
        :rtype: str
        """
        dbot_address = Web3.toChecksumAddress(dbot_address)
        w3 = self.channel_client.context.web3
        Dbot = _make_dbot_contract(w3, dbot_address)
        owner = Dbot.functions.getOwner().call()
        return owner

    def get_price(self, dbot_address: str, uri: str, method: str) -> int:
        """Get the price of a endpoint of the DBot

        The unit of price is `wei`, the smallest unit of ATN. 1ATN = 10^18wei

        :param dbot_address: address of the DBot contract
        :param uri: uri of the endpoint
        :param method: method of the endpoint
        :return: price of the DBot's endpoint
        :rtype: int
        """

        dbot_address = Web3.toChecksumAddress(dbot_address)
        w3 = self.channel_client.context.web3
        Dbot = _make_dbot_contract(w3, dbot_address)
        key = Dbot.functions.getKey(tobytes32(method.lower()), tobytes32(uri)).call()
        endpoint = Dbot.functions.keyToEndPoints(key).call()
        # TODO handle method case and how to check if the endpoint exist
        if (int(endpoint[1]) == 0):
            raise AtnException('no such endpoint: uri = {}, method = {}'.format(uri, method))
        else:
            return int(endpoint[1])

    def get_dbot_channel(self, dbot_address: str) -> Channel:
        """Get the channel information from DBot server

        DBot server saves the balance proof which send from DBot user.
        DBot users can get this from DBot server, no need to save by themselves.

        :param dbot_address: address of the DBot contract
        :return: :class:`Channel <Channel>` object
        :rtype: pyatn_client.Channel
        """
        dbot_address = Web3.toChecksumAddress(dbot_address)
        domain = self.get_dbot_domain(dbot_address)
        channel = self.get_channel(dbot_address)
        dbot_url = domain if domain.lower().startswith('http') else 'http://{}'.format(domain)
        if channel is not None:
            url = '{}/api/v1/dbots/{}/channels/{}/{}'.format(dbot_url,
                                                             channel.receiver,
                                                             channel.sender,
                                                             channel.block
                                                             )
            resp = requests.get(url)
            if resp.status_code == 200:
                return resp.json()
            else:
                return None
        else:
            return None

    def wait_dbot_sync(self, dbot_address: str, retry_interval: int=5, retry_times: int=5) -> None:
        """Wait the DBot server to sync the channel info on blockchain

        DBot server will sync the channel info on blockchain, and

        :param dbot_address: address of the DBot contract
        :param retry_interval: interval time for retry, seconds
        :param retry_times:  how many times to retry
        """
        dbot_address = Web3.toChecksumAddress(dbot_address)
        channel = self.get_channel(dbot_address)
        if channel is None:
            logger.warning('No Channel with dbot({}) on chain'.format(dbot_address))
            return
        dbot_channel = self.get_dbot_channel(dbot_address)
        remain_times = retry_times
        while retry_times > 0 and (dbot_channel is None or int(dbot_channel['deposit']) != channel.deposit):
            logger.info('Channel state with dbot({}) has not synced by dbot server, retry after {}s'.format(
                dbot_address, retry_interval))
            remain_times = remain_times - 1
            time.sleep(retry_interval)
            dbot_channel = self.get_dbot_channel(dbot_address)
        if dbot_channel is not None and int(dbot_channel['deposit']) == channel.deposit:
            channel.update_balance(int(dbot_channel['balance']))
        else:
            raise AtnException('Channel state with dbot({}) can not synced by dbot server.'.format(dbot_address))

    def call_dbot_api(self, dbot_address: str, uri: str, method: str, **requests_kwargs) -> Response:
        """Send the API's HTTP request

        Channel will be auto created if no channel or be topuped if
        insufficient balance in channel.
        The deposit value is determined by `deposit_strategy`.
        A signature of balance will be sent to DBot server to pay the price of the API.

        :param dbot_address: address of the DBot contract
        :param uri: uri of the endpoint
        :param method: method of the endpoint
        :param requests_kwargs: the other args for http request is same with `requests`
        :return: :class:`Response <Response>` object, http response of the API
        :rtype: requests.Response
        """
        dbot_address = Web3.toChecksumAddress(dbot_address)
        price = self.get_price(dbot_address, uri, method)
        channel = self._get_suitable_channel(dbot_address, price)
        channel.create_transfer(price)
        domain = self.get_dbot_domain(dbot_address)
        dbot_url = domain if domain.lower().startswith('http') else 'http://{}'.format(domain)
        url = '{}/call/{}/{}'.format(dbot_url, dbot_address, remove_slash_prefix(uri))
        return self._request(channel, method, url, **requests_kwargs)

    def open_channel(self, dbot_address: str, deposit: int) -> Channel:
        """Open a channel with the DBot

        If a channel with the DBot has exist, the channel will be return directlly.
        This function will block until the `ChannelCreated` event received
        from blockchain.

        :param dbot_address: address of the DBot contract
        :param deposit: the value of deposit
        :return: :class:`Channel <Channel>` object, None if failed.
        :rtype: Channel
        """
        channel = self.get_channel(dbot_address)
        if channel is not None:
            logger.warning('A channel is exist')
            return channel
        return self.channel_client.open_channel(dbot_address, deposit)

    def topup_channel(self, dbot_address: str, deposit: int) -> Channel:
        """Topup the channel with the DBot

        This function will block until the `ChannelToppedUp` event received
        from blockchain.

        :param dbot_address: address of the DBot contract
        :param deposit: the value of deposit
        :return: :class:`Channel <Channel>` object, `None` if failed.
        :rtype: Channel
        """
        return self.channel_client.topup_channel(dbot_address, deposit)

    def close_channel(self, dbot_address: str) -> None:
        """Close the channel with the DBot

        This function will block until the `ChannelSettled` event received
        from blockchain.
        If can not get valid close signature from DBot server,
        it will close the channel with balance 0 unilaterally.
        You can change the manual through redefine `on_cooperative_close_denied`

        :param dbot_address: address of the DBot contract
        """
        channel = self.get_channel(dbot_address)
        if channel is None:
            logger.error('No channel to close.')
            return

        try:
            self.wait_dbot_sync(dbot_address)
        except Exception as err:
            logger.error('Dbot server can not sync the channel')
            self.on_cooperative_close_denied(dbot_address, response)
        domain = self.get_dbot_domain(dbot_address)
        dbot_url = domain if domain.lower().startswith('http') else 'http://{}'.format(domain)

        logger.debug(
            'Requesting closing signature from server for balance {} on channel {}/{}/{}.'
            .format(
                channel.balance,
                channel.receiver,
                channel.sender,
                channel.block
            )
        )
        url = '{}/api/v1/channels/{}/{}/{}'.format(
            dbot_url,
            channel.receiver,
            channel.sender,
            channel.block
        )

        try:
            response = requests.request(
                'DELETE',
                url,
                params={'balance': channel.balance}
            )
        except requests.exceptions.ConnectionError as err:
            logger.error(
                'Could not get a response from the server while requesting a closing signature: {}'
                .format(err)
            )
            response = None

        failed = True
        if response is not None and response.status_code == requests.codes.OK:
            closing_sig = response.json()['close_signature']
            dbot_owner = self.get_dbot_owner(dbot_address)
            failed = channel.close_cooperatively(decode_hex(closing_sig), dbot_owner) is None

        if response is None or failed:
            logger.error('Cooperative close channel failed.')
            self.on_cooperative_close_denied(dbot_address, response)
        else:
            logger.info('Cooperative close channel successfully')

    def on_cooperative_close_denied(self, dbot_address: str, response: Response = None) -> None:
        """Call back function when no valid closing signature received

        This function will be called when can not get valid closing signature in
        method `close_channel`

        :param dbot_address: address of the DBot contract
        :param response: response from DBot server when request closing signature
        """
        logger.warning('No valid closing signature received from DBot server({}).\n{}'.format(dbot_address, response.text))
        logger.warning('Closing noncooperatively on a balance of 0.')
        # if cooperative close denied, client close the channel with balance 0 unilaterally
        self.uncooperative_close_channel(dbot_address, 0)

    def uncooperative_close_channel(self, dbot_address: str, balance: int) -> None:
        """Close the channel unilaterally

        If DBot server has down, you can close the channel with any balance.
        DBot server will close the channel if it's working and detect that
        you send a wrong balance.

        :param dbot_address: address of the DBot contract
        :param balance: used balance of the channel
        """
        channel = self.get_channel(dbot_address)
        if channel is None:
            logger.error('No channel to close.')
            return
        channel.close(balance)

    def settle_channel(self, dbot_address: str) -> None:
        """Settle the channel to withdraw your deposit after you close the channel unilaterally

        This function should be called after the end of the challenge peroid

        :param dbot_address: address of the DBot contract
        """
        channel = self.get_channel(dbot_address)
        if channel is None:
            logger.error('No channel to settle.')
            return
        channel.settle()

    def get_channel(self, dbot_address: str) -> Channel:
        """Get channel information from blockchain

        :param dbot_address: address of the DBot contract
        :return: :class:`Channel <Channel>` object, `None` is no channel
        :rtype: Channel
        """
        open_channels = self.channel_client.get_channels(dbot_address)
        if open_channels:
            channel = open_channels[0]
            return channel
        else:
            return None

    def _get_suitable_channel(self,
                             dbot_address: str,
                             price: int
                             ) -> Channel:
        if self.deposit_strategy is None:
            channel = self.get_channel(dbot_address)
            if channel is None:
                logger.error('No channel was found with DBot({}), please create a channel first'.format(dbot_address))
                raise AtnException('No channel was found with DBot({})'.format(dbot_address))
            if not channel.is_suitable(price):
                logger.error('Insufficient balance in the channel (remain balance = {}), please topup first'.format(
                    channel.remain_balance()))
                raise AtnException('Insufficient balance in the channel (remain balance = {}), please topup first'.format(
                    channel.remain_balance()))
            return channel
        else:
            channel = self.channel_client.get_suitable_channel(
                dbot_address, price, self.deposit_strategy, self.deposit_strategy
            )
            if channel is None:
                logger.error("No channel could be created or sufficiently topped up.")
                raise AtnException('No channel could be created or sufficiently topped up.')

            self.wait_dbot_sync(dbot_address)
            if channel.remain_balance() < price:
                channel.topup(self.deposit_strategy(price))
                self.wait_dbot_sync(dbot_address)
            return channel

    def _request(
            self,
            channel: Channel,
            method: str,
            url: str,
            **requests_kwargs
    ) -> Tuple[Union[None, Response], bool]:
        """
        Performs a simple request to the HTTP server with headers representing the given
        channel state.
        """
        headers = Munch()
        headers.contract_address = self.channel_client.context.channel_manager.address
        if channel is not None:
            headers.balance = str(channel.balance)
            headers.balance_signature = encode_hex(channel.balance_sig)
            headers.sender_address = channel.sender
            headers.receiver_address = channel.receiver
            headers.open_block = str(channel.block)

        headers = HTTPHeaders.serialize(headers)
        if 'headers' in requests_kwargs:
            headers.update(requests_kwargs['headers'])
            requests_kwargs['headers'] = headers
        else:
            requests_kwargs['headers'] = headers
        return requests.request(method, url, **requests_kwargs)



from typing import List, Any, Union, Dict

import time
from web3 import Web3
from web3.contract import Contract
from eth_account import Account

from ..config import NETWORK_CFG
from ..utils.populus_compat import LogFilter

DEFAULT_TIMEOUT = 60
DEFAULT_RETRY_INTERVAL = 3


def signed_transaction(
    web3: Web3,
    account: Account,
    to: str,
    value: int,
) -> str:

    tx = {
        'to': to,
        'value': value,
        'gasPrice': web3.eth.gasPrice,
        'nonce': web3.eth.getTransactionCount(account.address),
        'chainId': int(web3.version.network)
    }
    return account.signTransaction(tx)


def signed_contract_transaction(
    account: Account,
    contract: Contract,
    func_sig: str,
    args: List[Any],
    value: int=0
):
    web3 = contract.web3
    tx_data = contract.get_function_by_signature(func_sig)(*args).buildTransaction({
            'from': account.address,
            'nonce': web3.eth.getTransactionCount(account.address),
            'gasPrice': web3.eth.gasPrice,
            'value': value
        })
    return account.signTransaction(tx_data)


def get_logs(
        contract: Contract,
        event_name: str,
        from_block: Union[int, str] = 0,
        to_block: Union[int, str] = 'pending',
        argument_filters: Dict[str, Any] = None
):
    event_abi = [
        abi_element for abi_element in contract.abi
        if abi_element['type'] == 'event' and abi_element['name'] == event_name
    ]
    assert len(event_abi) == 1, 'No event found matching name {}.'.format(event_name)
    event_abi = event_abi[0]

    if argument_filters is None:
        argument_filters = {}

    tmp_filter = LogFilter(
        contract.web3,
        [event_abi],
        contract.address,
        event_name,
        from_block,
        to_block,
        argument_filters
    )
    logs = tmp_filter.get_logs()
    tmp_filter.uninstall()
    return logs


def _get_logs_raw(contract: Contract, filter_params: Dict[str, Any]):
    """For easy patching."""
    return contract.web3._requestManager.request_blocking('eth_getLogs', [filter_params])


def get_event_blocking(
        contract: Contract,
        event_name: str,
        from_block: Union[int, str] = 0,
        to_block: Union[int, str] = 'latest',
        argument_filters: Dict[str, Any]=None,
        condition=None,
        wait=DEFAULT_RETRY_INTERVAL,
        timeout=DEFAULT_TIMEOUT
) -> Union[Dict[str, Any], None]:
    for i in range(0, timeout + wait, wait):
        logs = get_logs(
            contract,
            event_name,
            from_block=from_block,
            to_block=to_block,
            argument_filters=argument_filters
        )
        matching_logs = [event for event in logs if not condition or condition(event)]
        if matching_logs:
            return matching_logs[0]
        elif i < timeout:
            _wait(wait)

    return None


def _wait(duration: float):
    """For easy patching."""
    time.sleep(duration)


def wait_for_transaction(
        web3: Web3,
        tx_hash: str,
        timeout: int = DEFAULT_TIMEOUT,
        polling_interval: int = DEFAULT_RETRY_INTERVAL
):
    for waited in range(0, timeout + polling_interval, polling_interval):
        tx_receipt = web3.eth.getTransactionReceipt(tx_hash)
        if tx_receipt is not None:
            return tx_receipt
        if waited < timeout:
            _wait(polling_interval)
    raise TimeoutError('Transaction {} was not mined.'.format(tx_hash))



import os
import json
from web3 import Web3

from ..constants import CONTRACT_METADATA, CHANNEL_MANAGER_NAME
from ..utils import privkey_to_addr


class Context(object):
    def __init__(
            self,
            private_key: str,
            web3: Web3,
            channel_manager_address: str
    ):
        self.private_key = private_key
        self.address = privkey_to_addr(private_key)
        self.web3 = web3
        self.account = web3.eth.account.privateKeyToAccount(private_key)

        self.channel_manager = web3.eth.contract(
            address=channel_manager_address,
            abi=CONTRACT_METADATA[CHANNEL_MANAGER_NAME]['abi']
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re
import json
import click
import importlib.util
import logging
import getpass
import requests
import time
from web3 import Web3, HTTPProvider
from web3.middleware import geth_poa_middleware

from eth_account import Account

from pyatn_client import Atn

def load_module(module_name, path):
    spec = importlib.util.spec_from_file_location(module_name, os.path.join(path, module_name + '.py'))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module

@click.group()
def cli():
    pass

@cli.command()
@click.option(
    '--output',
    default='keystore',
    help='Path to privatekey file and password file of the created account'
)
def create_account(output):
    """
    Create an encrypted account on ATN test net.
    """
    acct = Account.create()
    password = getpass.getpass("Enter the password to encrypt your account: ")
    encrypted = Account.encrypt(acct.privateKey, password)
    os.makedirs(output, exist_ok=True)
    acct_path = os.path.join(output, acct.address)
    os.makedirs(acct_path, exist_ok=True)
    pk_file = os.path.join(acct_path, 'privatekey')
    pw_file = os.path.join(acct_path, 'password')
    with open(pk_file, 'w') as fh:
        fh.write(json.dumps(encrypted, indent=2))
    with open(pw_file, 'w') as fh:
        fh.write(password)
    click.echo('A new account {} created.'.format(acct.address))
    click.echo('The private  file saved at {}'.format(os.path.abspath(pk_file)))
    click.echo('The password file saved at {}'.format(os.path.abspath(pw_file)))

@cli.command()
@click.option(
    '--address',
    required=True,
    help='Address of the account to request ATN'
)
def get_atn(address):
    """
    Get ATN from the faucet server of ATN test net
    """
    w3 = Web3(HTTPProvider('https://rpc-test.atnio.net'))
    w3.middleware_stack.inject(geth_poa_middleware, layer=0)
    click.echo('ATN Balance is: {} ATN'.format(w3.fromWei(w3.eth.getBalance(address), 'ether')))
    resp = requests.post('http://119.3.57.66:4111/faucet/{}'.format(address))
    if resp.status_code == 200:
        click.echo('Get 100 ATN successfully. (One address can get 100 ATN everyday.)')
        click.echo('Waiting 10 second for the transaction finished ...')
        for i in range(10, 0, -1):
            time.sleep(1)
            click.echo(i)
        click.echo('ATN Balance is: {} ATN'.format(w3.fromWei(w3.eth.getBalance(address), 'ether')))
    else:
        click.echo('Can not get ATN. Try to visit out faucet page "https://faucet-test.atnio.net/"')

@cli.command()
@click.option(
    '--address',
    required=True,
    help='Address of the account to request ATN'
)
def get_balance(address):
    click.echo('ATN Balance is: {} ATN'.format(w3.fromWei(w3.eth.getBalance(address), 'ether')))

@cli.command()
@click.option(
    '--pk-file',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
    help='Path to private key file'
)
@click.option(
    '--pw-file',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
    help='Path to file containing the password for the private key specified.'
)
@click.option(
    '--http-provider',
    default='https://rpc-test.atnio.net',
    help='Web3 Http Provider, default is https://rpc-test.atnio.net'
)
@click.option(
    '--dbot-address',
    required=True,
    help='Address of the DBot contract'
)
@click.option(
    '--data',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
    help="Path to python file containing requests data for the DBot API"
)
def call(
        pk_file: str,
        pw_file: str,
        http_provider: str,
        dbot_address: str,
        data: str
):
    """
    Call an API of the DBot.

    """

    requests_test = load_module(os.path.splitext(os.path.basename(data))[0],
                                os.path.dirname(os.path.abspath(data)))
    requests_data = requests_test.data

    atn = Atn(
        http_provider=http_provider,
        pk_file=pk_file,
        pw_file=pw_file
    )

    #  channel = atn.get_suitable_channel(dbot_address, requests_data['endpoint']['uri'], requests_data['endpoint']['method'])
    response = atn.call_dbot_api(dbot_address,
                                 uri=requests_data['endpoint']['uri'],
                                 method=requests_data['endpoint']['method'],
                                 **requests_data['kwargs'])

    if response.status_code == 200:
        click.echo('Got 200 Response. Content-Type: {}'.format(response.headers['Content-Type']))
        if re.match('^json/', response.headers['Content-Type']):
            click.echo(response.headers)
            click.echo(response.json)
        elif response.headers['Content-Type'].startswith('audio') or response.headers['Content-Type'].startswith('image'):
            click.echo(response.headers)
            response_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'response_file')
            with open(response_file, 'wb') as f:
                f.write(response.content)
            click.echo('Save response file at {}'.format(response_file))
        else:
            click.echo(response.headers)
            click.echo(response.text)
    else:
        click.echo('Got {} Response.'.format(response.status_code))
        click.echo(response.text)


@cli.command()
@click.option(
    '--pk-file',
    required=True,
    help='Path to private key file or a hex-encoded private key.'
)
@click.option(
    '--pw-file',
    default=None,
    help='Path to file containing the password for the private key specified.',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    '--http-provider',
    default='https://rpc-test.atnio.net',
    help='Web3 Http Provider, default is https://rpc-test.atnio.net'
)
@click.option('--dbot-address', required=True, help='dbot address')
def close(
    pk_file: str,
    pw_file: str,
    http_provider: str,
    dbot_address: str,
):
    """
    Close the channel with a DBot.
    """
    atn = Atn(
        http_provider=http_provider,
        pk_file=pk_file,
        pw_file=pw_file
    )

    atn.close_channel(dbot_address)

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    cli()



"""This module contains network-specific defaults for different networks.
You can change i.e. gas price, gas limits, or contract address here.

Example:
    Set global network defaults for ropsten::

        from config import NETWORK_CFG
        from constants import get_network_id

        NETWORK_CFG.set_defaults(get_network_id('ropsten'))

    Change global gas price::
    from config import NETWORK_CFG
        NETWORK_CFG.gas_price = 15 * denoms.gwei
"""
from eth_utils import denoms
from collections import namedtuple, OrderedDict
from functools import partial

# these are default values for network config
network_config_defaults = OrderedDict(
    (('channel_manager_address', None),
     ('start_sync_block', 0),
     ('gas_price', 20 * denoms.gwei),
     ('gas_limit', 130000),
     # pot = plain old transaction, for lack of better term
     ('pot_gas_limit', 21000))
)
# create network config type that supports defaults
NetworkConfig = partial(
    namedtuple(
        'NetworkConfig',
        network_config_defaults
    ),
    **network_config_defaults
)

# network-specific configuration
NETWORK_CONFIG_DEFAULTS = {
    # testnet
    17: NetworkConfig(
        channel_manager_address='0x0000000000000000000000000000000000000012',
        start_sync_block=0
    )
}

class NetworkRuntime:
    def __init__(self):
        super().__setattr__('cfg', None)

    def set_defaults(self, network_id: int):
        """Set global default settings for a given network id.

        Args:
            network_id (int): a network id to use.
        """
        cfg_copy = dict(NETWORK_CONFIG_DEFAULTS[network_id]._asdict())
        super().__setattr__('cfg', cfg_copy)

    def __getattr__(self, attr):
        return self.cfg.__getitem__(attr.lower())

    def __setattr__(self, attr, value):
        if attr == 'cfg':
            return super().__setattr__('cfg', value)
        return self.cfg.__setitem__(attr.lower(), value)


def get_defaults(network_id: int):
    return NETWORK_CONFIG_DEFAULTS[network_id]


# default config
NETWORK_CFG = NetworkRuntime()
NETWORK_CFG.set_defaults(17)



import inspect
from typing import Any, Dict, Callable


def get_function_kwargs(kwargs: Dict[str, Any], function: Callable):
    return {
        kw: arg for kw, arg in kwargs.items()
        if kw in inspect.signature(function).parameters
    }


def pop_function_kwargs(kwargs: Dict[str, Any], function: Callable):
    function_kwargs = get_function_kwargs(kwargs, function)
    for kw in function_kwargs:
        kwargs.pop(kw)

    return function_kwargs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import logging
from logging.handlers import RotatingFileHandler

LOGPATH = os.path.join(os.path.dirname(__file__), 'logs')
os.makedirs(LOGPATH, exist_ok=True)

class AtnLogger():
    def __init__(self, name: str):
        self._name = name

    def config(self):
        return {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'verbose': {
                    'format': '%(asctime)s %(name)s(%(levelname)s) %(message)s (%(filename)s[%(lineno)d])'
                },
                'normal': {
                    'format': '%(asctime)s %(levelname)-8s %(message)s'
                },
                'simple': {
                    'format': '%(levelname)s %(message)s'
                },
            },
            'handlers': {
                'console':{
                    'level': 'INFO',
                    'class':'logging.StreamHandler',
                    'formatter': 'simple'
                },
                'info_file': {
                    'level': 'INFO',
                    'class': 'logging.handlers.RotatingFileHandler',
                    'formatter': 'normal',
                    'filename': os.path.join(LOGPATH, '{}.log'.format(self._name)),
                    'encoding': 'utf8',
                    'mode': 'a',
                    'maxBytes': 10485760,
                    'backupCount': 5
                },
                'debug_file': {
                    'level': 'DEBUG',
                    'class': 'logging.handlers.RotatingFileHandler',
                    'formatter': 'verbose',
                    'filename': os.path.join(LOGPATH, '{}_debug.log'.format(self._name)),
                    'encoding': 'utf8',
                    'mode': 'a',
                    'maxBytes': 10485760,
                    'backupCount': 20
                },
            },
            'loggers': {
                'atn': {
                    'handlers': ['console', 'info_file', 'debug_file'],
                    'level': 'DEBUG',
                    'propagate': False
                }
            },
            'root': {
                'handlers': ['console', 'info_file', 'debug_file'],
                'level': 'DEBUG',
            }
        }



import os
import logging
from typing import List

import os
from eth_utils import decode_hex, is_same_address, is_hex, remove_0x_prefix, to_checksum_address
from web3 import Web3
from web3.providers.rpc import HTTPProvider
from web3.utils.filters import Filter

from ..utils import (
    get_private_key,
    get_logs,
    get_event_blocking,
    signed_contract_transaction,
    signed_contract_transaction
)

from ..config import NETWORK_CFG
from ..constants import WEB3_PROVIDER_DEFAULT
from .context import Context
from .channel import Channel

logger = logging.getLogger('atn.microraiden' + os.path.splitext(os.path.basename(__file__))[0])


class Client:
    def __init__(
            self,
            private_key: str = None,
            key_password_path: str = None,
            channel_manager_address: str = None,
            web3: Web3 = None
    ) -> None:
        is_hex_key = is_hex(private_key) and len(remove_0x_prefix(private_key)) == 64
        is_path = os.path.exists(private_key)
        assert is_hex_key or is_path, 'Private key must either be a hex key or a file path.'

        # Load private key from file if none is specified on command line.
        if is_path:
            private_key = get_private_key(private_key, key_password_path)
            assert private_key is not None, 'Could not load private key from file.'

        self.channels = []  # type: List[Channel]

        # Create web3 context if none is provided, either by using the proxies' context or creating
        # a new one.
        if not web3:
            web3 = Web3(HTTPProvider(WEB3_PROVIDER_DEFAULT))

        channel_manager_address = to_checksum_address(
            channel_manager_address or NETWORK_CFG.CHANNEL_MANAGER_ADDRESS
        )

        self.context = Context(private_key, web3, channel_manager_address)

        self.sync_channels()

    def sync_channels(self):
        filter_kwargs = {
            'fromBlock': 0,
            'toBlock': 'latest'
        }
        filters = {'_sender_address': self.context.address}
        create = self.context.channel_manager.events.ChannelCreated.createFilter(
            argument_filters=filters,
            **filter_kwargs).get_all_entries()

        topup = self.context.channel_manager.events.ChannelToppedUp.createFilter(
            argument_filters=filters,
            **filter_kwargs).get_all_entries()

        close = self.context.channel_manager.events.ChannelCloseRequested.createFilter(
            argument_filters=filters,
            **filter_kwargs).get_all_entries()

        settle = self.context.channel_manager.events.ChannelSettled.createFilter(
            argument_filters=filters,
            **filter_kwargs).get_all_entries()

        channel_key_to_channel = {}

        def get_channel(event) -> Channel:
            sender = to_checksum_address(event['args']['_sender_address'])
            receiver = to_checksum_address(event['args']['_receiver_address'])
            block = event['args'].get('_open_block_number', event['blockNumber'])
            assert is_same_address(sender, self.context.address)
            return channel_key_to_channel.get((sender, receiver, block), None)

        for c in self.channels:
            channel_key_to_channel[(c.sender, c.receiver, c.block)] = c

        for e in create:
            c = get_channel(e)
            if c:
                c.deposit = e['args']['_deposit']
            else:
                c = Channel(
                    self.context,
                    to_checksum_address(e['args']['_sender_address']),
                    to_checksum_address(e['args']['_receiver_address']),
                    e['blockNumber'],
                    e['args']['_deposit'],
                    on_settle=lambda channel: self.channels.remove(channel)
                )
                assert is_same_address(c.sender, self.context.address)
                channel_key_to_channel[(c.sender, c.receiver, c.block)] = c

        for e in topup:
            c = get_channel(e)
            c.deposit += e['args']['_added_deposit']

        for e in close:
            # Requested closed, not actual closed.
            c = get_channel(e)

            c.update_balance(e['args']['_balance'])
            c.state = Channel.State.settling

        for e in settle:
            c = get_channel(e)
            c.state = Channel.State.closed

        # Forget closed channels.
        self.channels = [
            c for c in channel_key_to_channel.values() if c.state != Channel.State.closed
        ]

        logger.debug('Synced a total of {} channels.'.format(len(self.channels)))

    def open_channel(self, receiver_address: str, deposit: int):
        """
        Attempts to open a new channel to the receiver with the given deposit. Blocks until the
        creation transaction is found in a pending block or timeout is reached. The new channel
        state is returned.
        """
        assert isinstance(receiver_address, str)
        assert isinstance(deposit, int)
        assert deposit > 0

        balance = self.context.web3.eth.getBalance(self.context.address)
        #  token_balance = self.context.token.call().balanceOf(self.context.address)
        if balance < deposit:
            logger.error(
                'Insufficient balance available for the specified deposit ({}/{})'
                .format(balance, deposit)
            )
            return None

        current_block = self.context.web3.eth.blockNumber
        logger.info('Creating channel to {} with an initial deposit of {} @{}'.format(
            receiver_address, deposit, current_block
        ))

        tx = signed_contract_transaction(
            self.context.account,
            self.context.channel_manager,
            'createChannel(address)',
            [
                receiver_address
            ],
            deposit
        )
        ret = self.context.web3.eth.sendRawTransaction(tx.rawTransaction)
        logger.info('transaction hash: {}'.format(Web3.toHex(ret)))

        logger.debug('Waiting for channel creation event on the blockchain...')
        filters = {
            '_sender_address': self.context.address,
            '_receiver_address': receiver_address
        }
        event = get_event_blocking(
            self.context.channel_manager,
            'ChannelCreated',
            from_block=current_block + 1,
            to_block='latest',
            argument_filters=filters
        )

        if event:
            logger.debug('Event received. Channel created in block {}.'.format(event['blockNumber']))
            assert is_same_address(event['args']['_sender_address'], self.context.address)
            assert is_same_address(event['args']['_receiver_address'], receiver_address)
            channel = Channel(
                self.context,
                self.context.address,
                receiver_address,
                event['blockNumber'],
                event['args']['_deposit'],
                on_settle=lambda c: self.channels.remove(c)
            )
            self.channels.append(channel)
        else:
            logger.error('Error: No event received.')
            channel = None

        return channel

    def topup_channel(self, receiver_address: str, deposit: int):
        open_channels = self.get_open_channels(receiver_address)
        if open_channels:
            capacity, channel = max(
                ((c.deposit - c.balance, c) for c in open_channels), key=lambda x: x[0]
            )
            event = channel.topup(deposit)
            return channel if event else None
        else:
            logger.error('Error: No channel to top up')
            return None

    def get_channels(self, receiver: str = None) -> List[Channel]:
        return [
            c for c in self.channels
            if is_same_address(c.sender, self.context.address) and
            (not receiver or is_same_address(c.receiver, receiver))
        ]

    def get_open_channels(self, receiver: str = None) -> List[Channel]:
        """
        Returns all open channels to the given receiver. If no receiver is specified, all open
        channels are returned.
        """
        return [
            c for c in self.channels
            if is_same_address(c.sender, self.context.address) and
            (not receiver or is_same_address(c.receiver, receiver)) and
            c.state == Channel.State.open
        ]

    def get_suitable_channel(
            self, receiver, value, initial_deposit=None, topup_deposit=None
    ) -> Channel:
        """
        Searches stored channels for one that can sustain the given transfer value. If none is
        found, a possibly open channel is topped up using the topup callable to determine its topup
        value. If both attempts fail, a new channel is created based on the initial deposit
        callable.
        Note: In the realistic case that only one channel is opened per (sender,receiver) pair,
        this method usually performs like this:
        1. Directly return open channel if sufficiently funded.
        2. Topup existing open channel if insufficiently funded.
        3. Create new channel if no open channel exists.
        If topping up or creating fails, this method returns None.
        Channels are topped up just enough so that their remaining capacity equals
        topup_deposit(value).
        """
        open_channels = self.get_open_channels(receiver)
        suitable_channels = [c for c in open_channels if c.is_suitable(value)]

        if suitable_channels:
            # At least one channel with sufficient funds.

            if len(suitable_channels) > 1:
                # This is not impossible but should only happen after bad channel management.
                logger.warning(
                    'Warning: {} suitable channels found. '
                    'Choosing the one with the lowest remaining capacity.'
                    .format(len(suitable_channels))
                )

            capacity, channel = min(
                ((c.deposit - c.balance, c) for c in suitable_channels), key=lambda x: x[0]
            )
            logger.debug(
                'Found suitable open channel, opened at block #{}.'.format(channel.block)
            )
            return channel

        elif open_channels:
            # Open channel(s) but insufficient funds. Requires topup.

            if len(open_channels) > 1:
                # This is not impossible but should only happen after bad channel management.
                logger.warning(
                    'Warning: {} open channels for topup found. '
                    'Choosing the one with the highest remaining capacity.'
                    .format(len(open_channels))
                )

            capacity, channel = max(
                ((c.deposit - c.balance, c) for c in open_channels), key=lambda x: x[0]
            )
            deposit = max(value, topup_deposit(value)) - capacity
            event = channel.topup(deposit)
            return channel if event else None

        else:
            # No open channels to receiver. Create a new one.
            deposit = max(value, initial_deposit(value))
            return self.open_channel(receiver, deposit)
